"""
Copyright (c) 2018 DecisionVis, LLC. All rights reserved.

Redistribution and use in source and binary forms, with
or without modification, are permitted provided that the
following conditions are met:

1. Redistributions of source code must retain the above
copyright notice, this list of conditions and the following
disclaimer.

2. Redistributions in binary form must reproduce the
above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other
materials provided with the distribution.

3. Neither the name of the copyright holder nor the names
of its contributors may be used to endorse or promote
products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

__all__ = ["Constants", "Structures", "Functions"]

from .Constants import MAXIMIZE
from .Constants import MINIMIZE
from .Constants import CENTERPOINT
from .Constants import OFAT
from .Constants import CORNERS
from .Constants import RANDOM
from .Constants import COUNT
from .Constants import RETAIN
from .Constants import DISCARD

from .Structures import Decision
from .Structures import Objective
from .Structures import Constraint
from .Structures import Tagalong
from .Structures import Problem
from .Structures import ArchiveIndividual
from .Structures import Individual
from .Structures import Rank
from .Structures import MOEAState

from .Functions import create_moea_state
from .Functions import doe
from .Functions import return_evaluated_individual
from .Functions import get_sample
from .Functions import get_iterator
from .Functions import decisions_to_grid_point

from .Sampling import NearExhaustionWarning
from .Sampling import TotalExhaustionError




"""
Copyright (c) 2018 DecisionVis, LLC. All rights reserved.

Redistribution and use in source and binary forms, with
or without modification, are permitted provided that the
following conditions are met:

1. Redistributions of source code must retain the above
copyright notice, this list of conditions and the following
disclaimer.

2. Redistributions in binary form must reproduce the
above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other
materials provided with the distribution.

3. Neither the name of the copyright holder nor the names
of its contributors may be used to endorse or promote
products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

# DOEState stages
from .Constants import CENTERPOINT
from .Constants import OFAT
from .Constants import CORNERS
from .Constants import RANDOM
from .Constants import EXHAUSTIVE
from .Constants import EXHAUSTED

from math import floor
from math import ceil

class NearExhaustionWarning(Exception):
    def __init__(self, state, *args, **kwargs):
        super(NearExhaustionWarning, self).__init__(*args, **kwargs)
        self.state = state

class TotalExhaustionError(Exception):
    def __init__(self, state, *args, **kwargs):
        super(TotalExhaustionError, self).__init__(*args, **kwargs)
        self.state = state

def is_duplicate(state, grid_point):
    if grid_point in state.issued.issued_set:
        return True
    if grid_point in state.archive_set:
        return True
    #for rank in state.archive:
    #    counter = 0
    #    for arch_ind in rank.individuals:
    #        if counter >= rank.occupancy:
    #            break
    #        if arch_ind.valid:
    #            counter += 1
    #            if grid_point == arch_ind.grid_point:
    #                return True
    return False

def doe_next(state):
    """
    state (MOEAstate)

    Return the next DOE grid point and update state accordingly.
    Will raise an exception if random sampling is having
    trouble finding a free grid point.
    Will raise another exception if a full grid sweep is
    unable to find a free grid point.  Users would be well
    advised to stop after the first exception!  After the
    second exception, we have to start duplicating samples.
    """
    doestate = state.doestate
    stage = doestate.stage
    grid = state.grid
    duplicated = True
    duplicates_generated = 0
    while duplicated:
        if stage == CENTERPOINT:
            grid_point = grid.GridPoint(*(len(a) // 2 for a in grid.axes))
            doestate = doestate._replace(stage=OFAT, counter=0)
        elif stage == OFAT:
            indices = [len(a) // 2 for a in grid.axes]
            counter = doestate.counter
            axis_index = counter // 2 # two points per axis
            axis = grid.axes[axis_index]
            if counter % 2 == 0:
                # low side sample
                indices[axis_index] = 0
            else:
                # high side sample
                indices[axis_index] = len(axis) - 1
            grid_point = grid.GridPoint(*indices)
            advanced_counter = counter + 1
            if advanced_counter // 2 >= len(grid.axes):
                doestate = doestate._replace(stage=RANDOM, counter=0)
            else:
                doestate = doestate._replace(counter=advanced_counter)
        elif stage == CORNERS:
            # Treat the counter like a bitvector, where 0 means low and 1
            # means high.  Because it's easy to write this way, treat
            # axis 0 as the LSB.
            indices = list()
            for ii, axis in enumerate(grid.axes):
                bit = (doestate.counter >> ii) & 1
                if bit == 0:
                    indices.append(0)
                else:
                    indices.append(len(axis) - 1)
            grid_point = grid.GridPoint(*indices)
            advanced_counter = doestate.counter + 1
            if advanced_counter >= 2 ** len(grid.axes):
                doestate = doestate._replace(stage=CENTERPOINT, counter=0)
            else:
                doestate = doestate._replace(counter=advanced_counter)
        elif stage in (RANDOM, EXHAUSTED):
            grid_point = grid.GridPoint(
                *(state.randint(0, len(a) - 1) for a in grid.axes))
            # transition out of RANDOM is if we have too many failures
            doestate = doestate._replace(counter=doestate.counter + 1)
        elif stage == EXHAUSTIVE:
            # Decompose the counter into indices, again treating axis
            # 0 as the least significant because it's easy to write.
            # This also takes advantage of the fact that Python has
            # bignums.  In C we need to check overflow,
            # although if we overflow 2 ** 63 - 1, that means
            # we've sampled almost 1e19 grid points.  So good
            # for us, if we trip over that condition!
            indices = list()
            counter = doestate.counter
            total_points = 1
            for axis in grid.axes:
                alen = len(axis)
                total_points = total_points * alen # mind overflow here!
                indices.append(counter % alen)
                counter = counter // alen
            advanced_counter = doestate.counter + 1
            if advanced_counter >= total_points:
                doestate = doestate._replace(stage=EXHAUSTED, counter=0)
                state = state._replace(doestate=doestate)
                raise TotalExhaustionError(state)
            else:
                doestate = doestate._replace(counter=advanced_counter)
        if stage == EXHAUSTED:
            # No point in wasting time on a duplicate check, and
            # if we're exhausted the loop condition will remain true
            # forever.
            break
        duplicated = is_duplicate(state, grid_point)
        if duplicated and stage == RANDOM:
            duplicates_generated += 1
            # How hard do we want to try here?  This says if the space
            # is more than 99.9% sampled on average we should move over
            # to exhaustive search.
            if duplicates_generated > 1000:
                doestate = doestate._replace(stage=EXHAUSTIVE)
                state = state._replace(doestate=doestate)
                raise NearExhaustionWarning(state)
    state = state._replace(doestate=doestate)
    return state, grid_point

def evolve(state):
    """
    state (MOEAState)

    This function produces a grid point by performing selection and
    variation.  It is the primary means by which we search for
    superior individuals in the problem space.
    """
    total_archive_occupancy = sum(r.occupancy for r in state.archive)
    # if archive is too small, return doe_next
    if total_archive_occupancy < 2:
        return doe_next(state)
    randint = state.randint

    # Ramp is greater than -1, and it determines how likely each rank
    # is to be selected for Parent B.
    ramp = -1

    duplicated = True
    circuit_breaker = 0
    while duplicated and circuit_breaker < 10:
        # Select parent A from rank 0.  Parents A and B are grid points.
        parent_a = _select(state, 0)
        if randint(1, 10) == 1:
            # do continuous injection
            draw = randint(0, 99)
            breaks = (10,30,60,80,90,95,100)
            n_dv_to_modify = 1
            for ii, limit in enumerate(breaks):
                count = ii + 1
                if draw >= limit or count == len(parent_a):
                    n_dv_to_modify = count
                    break
            offspring = parent_a
            while n_dv_to_modify > 0:
                n_dv_to_modify -= 1
                dv_index = randint(0, len(parent_a) - 1)
                while offspring[dv_index] != parent_a[dv_index]:
                    # don't repeat yourself
                    dv_index = randint(0, len(parent_a) - 1)
                axis = state.grid.axes[dv_index]
                field = offspring._fields[dv_index]
                # uniform random mutation, different sense
                # of the word "index"
                new_index = randint(0, len(axis) - 2)
                # cut a hole for the current index
                # (hence the -2 above)
                if new_index >= offspring[dv_index]:
                    new_index += 1
                offspring = offspring._replace(**{field: new_index})
        else:
            # do SBX
            parent_b = parent_a
            while parent_b == parent_a:
                # Choose a rank for parent B.  Parent B's role is not symmetrical
                # with parent A, unlike in traditional SBX.  This selection-and-
                # variation procedure uses one or two variables from Parent B to mutate
                # Parent A, so the offspring distribution is centered on Parent A.
                # This is a lot more like a mutation operator with a 1/ndv rate
                # than a traditional crossover operator.  But like DE, it is using
                # the population to determine an appropriate step size.
                rank_b = _select_rank(state, ramp)
                ramp += 1
                # Choose parent B from its rank.
                parent_b = _select(state, rank_b)

            # Find unequal decision variables.  (C99 allows variable-length
            # arrays.  It's a stack allocation, but it would take a _huge_
            # decision space to blow out the stack.  Still, it's a risk.
            # The alternative would be to malloc a few extra index arrays
            # into State.)
            dv_equality = [aa == bb for aa, bb in zip(parent_a, parent_b)]
            n_unequal_dvs = len(dv_equality) - sum(dv_equality)

            # Choose how many DVs to mutate
            # Here's a performance distribution based on a weak sparsity
            # of effects assumption.
            # 10% 1 variable 10
            # 20% 2 variable 30
            # 30% 3 variable 60
            # 20% 4 variable 80
            # 10% 5 variable 90
            # 5%  6 variable 95
            # 5%  7 variable 100
            draw = randint(0, 99)
            breaks = (10,30,60,80,90,95,100)

            n_dv_to_modify = 1
            for ii, limit in enumerate(breaks):
                count = ii + 1
                if n_unequal_dvs <= count or draw < limit:
                    n_dv_to_modify = count
                    break

            # Apply SBX 
            offspring = parent_a
            previous_dv = -1
            while n_dv_to_modify > 0:
                # randomly select from the unequal decision variables
                target = randint(0, n_unequal_dvs - 1)
                target_index = None
                counter = 0
                for index, isequal in enumerate(dv_equality):
                    if not isequal:
                        if counter == target:
                            target_index = index
                            break
                        counter += 1

                # Do SBX on the chosen variable, in index space.
                result = sbx_index(
                    offspring[target_index],
                    parent_b[target_index],
                    len(state.grid.axes[target_index]),
                    state.random)

                # Look up the name of the field so we can do _replace
                field = offspring._fields[target_index]
                offspring = offspring._replace(**{field: result})

                # Prepare for next iteration
                n_dv_to_modify -= 1
                # Adjust dv_equality so that we don't hit the same DV again.
                dv_equality[target_index] = True
                n_unequal_dvs -= 1

        # Treat the offspring as a direction for a line search.
        # If the offspring is not duplicated, we'll just get it back.
        offspring, duplicated = _line_search(state, parent_a, offspring)

        # Increment the "circuit breaker" so that we don't line search
        # forever in a saturated space
        circuit_breaker += 1

    # If everything failed, return a doe point
    if duplicated:
        return doe_next(state)
    return state, offspring

def sbx_index(aa, bb, allowed, random):
    """
    Perform SBX on indices.  Return a new index.

    TODO: allow to parameterize DI.

    aa (int >= 0): an index
    bb (int >= 0): an index
    allowed (int > 1): number of allowed indices
    """
    if allowed == 1:
        raise Exception("This operation will always return 0!")
    rounded_result = aa
    while rounded_result == aa:
        result = sbx(0.0, float(allowed-1), float(aa), float(bb), 1.0, random)
        difference = result - aa
        if difference > 0:
            rounded_result = aa + int(ceil(difference))
            if rounded_result >= allowed:
                rounded_result = allowed - 1
        elif difference < 0:
            rounded_result = aa + int(floor(difference))
            if rounded_result < 0:
                rounded_result = 0
    return rounded_result

def sbx(x_lower, x_upper, x_parent1, x_parent2, di, random):
    """
    Perform SBX on one variable.

    x_lower (float): lower bound on DV
    x_upper (float): upper bound on DV
    x_parent1 (float): primary parent value
    x_parent2 (float): secondary parent value
    di (float, nonnegative): distribution index
    """
    x_range = x_upper - x_lower
    gamma = di + 1.0
    kappa = 1.0 / gamma
    if x_parent1 == x_parent2:
        # early return if parents are the same
        return x_parent1
    swapped = False
    if x_parent1 < x_parent2:
        y_1 = (x_parent1 - x_lower) / x_range
        y_2 = (x_parent2 - x_lower) / x_range
    else:
        swapped = True
        y_2 = (x_parent1 - x_lower) / x_range
        y_1 = (x_parent2 - x_lower) / x_range
    delta = y_2 - y_1
    beta = 1.0 + (2.0/delta) * min(y_1, 1-y_2)
    alpha = 2.0 - beta ** (-gamma)
    uniform_1 = random()
    if uniform_1 <= 1.0 / alpha:
        beta_q = (uniform_1 * alpha) ** kappa
    else:
        beta_q = (1.0 / (2.0 - uniform_1 * alpha)) ** kappa
        # This assertion is not true! assert(beta_q <= 1.0)
    if swapped:
        sign = -1.0
    else:
        sign = 1.0
    y_child = 0.5 * ((y_1 + y_2) + sign * beta_q * delta)
    x_child = x_lower + x_range * y_child
    return x_child

def _line_search(state, parent, offspring):
    """
    state (MOEAState)
    parent (GridPoint)
    offspring (GridPoint)

    Do a line search in the parent->offspring direction

    Return a grid point and whether or not that point is duplicated.
    (grid_point, duplicated)
    """
    if not is_duplicate(state, offspring):
        return offspring, False
    # See comment in evolve() about stack allocation versus heap
    # allocation for "working" indices and flags.
    step = [o - p for o, p in zip(offspring, parent)]
    abstep = [abs(s) for s in step]
    signstep = [(int(s >= 0) * 2 - 1) * int(s != 0) for s in step]
    threshold = max(abstep)
    counters = [0 for _ in abstep]
    duplicated = True
    location = [o for o in offspring]
    search_result = offspring
    # Search further out from offspring
    failed = False
    while duplicated and not failed:
        failed = False
        for ii in range(len(counters)):
            counters[ii] += abstep[ii]
            if counters[ii] >= threshold:
                counters[ii] = counters[ii] % threshold
                location[ii] = location[ii] + signstep[ii]
                if location[ii] < 0:
                    failed = True
                elif location[ii] >= len(state.grid.axes[ii]):
                    failed = True
        search_result = state.grid.GridPoint(*location)
        duplicated = is_duplicate(state, search_result)
    if duplicated or failed:
        # Search toward parent from offspring
        location = [o for o in offspring]
        counters = [0 for _ in abstep]
        search_result = offspring
        failed = False
        duplicated = True
        while duplicated and not failed:
            failed = False
            for ii in range(len(counters)):
                counters[ii] += abstep[ii]
                if counters[ii] >= threshold:
                    counters[ii] = counters[ii] % threshold
                    location[ii] = location[ii] - signstep[ii]
                    if location[ii] < 0:
                        failed = True
                    elif location[ii] >= len(state.grid.axes[ii]):
                        failed = True
            search_result = state.grid.GridPoint(*location)
            duplicated = is_duplicate(state, search_result)

    if failed or duplicated:
        return offspring, True
    else:
        return search_result, False

def _select_rank(state, ramp):
    """
    ramp (float): Weighting ramp for relative rank probability.
                  Ramp is greater than or equal to -1.

    Returns the index of a rank to sample.
    """
    if -1.0 <= ramp:
        pass # in the domain
    else:
        raise Exception("Ramp {} is not greater than -1".format(ramp))
    occupied_ranks = 0
    for rank in state.archive:
        if rank.occupancy == 0:
            break
        occupied_ranks += 1
    # see elsewhere discussion about stack allocation
    intervals = [occupied_ranks + ramp * ii for ii in range(occupied_ranks)]
    breaks = [sum(intervals[:(ii+1)]) for ii in range(occupied_ranks)]
    limit = breaks[-1]
    selection = state.randint(0, limit)
    rank_number = 0
    while breaks[rank_number] < selection:
        rank_number += 1
    return rank_number

def _select(state, rank_number):
    """
    rank_number (index):  An index into the archive.

    Returns a GridPoint from the given rank.
    If the rank is empty, raises an Exception.
    If the rank number is outside the archive, raises
    an IndexError.  Don't do that!
    """
    rank = state.archive[rank_number]
    if rank.occupancy == 0:
        raise Exception("Can't select from an empty rank ({}).".format(rank_number))
    randint = state.randint
    target = randint(0, rank.occupancy - 1)
    counter = -1
    index = -1
    while counter < target:
        index += 1
        if rank.individuals[index].valid:
            counter += 1
        if index > len(rank.individuals):
            # Should never get here because occupancy is the number
            # of valid individuals in the rank.  If we do, it's because
            # of a bookkeeping failure somewhere else.  Probably in
            # move_individual.  Be very careful about updating the
            # contents of rank.individuals, because it's mutable!  Old
            # Ranks will be wrong about occupancy.
            raise Exception("Rank {} is inconsistent!".format(rank_number))
    return rank.individuals[index].grid_point




"""
Copyright (c) 2018 DecisionVis, LLC. All rights reserved.

Redistribution and use in source and binary forms, with
or without modification, are permitted provided that the
following conditions are met:

1. Redistributions of source code must retain the above
copyright notice, this list of conditions and the following
disclaimer.

2. Redistributions in binary form must reproduce the
above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other
materials provided with the distribution.

3. Neither the name of the copyright holder nor the names
of its contributors may be used to endorse or promote
products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

from collections import namedtuple

from math import floor

from random import random
from random import randint

from .Constants import MAXIMIZE
from .Constants import MINIMIZE

from .Constants import CENTERPOINT
from .Constants import OFAT
from .Constants import CORNERS
from .Constants import RANDOM
from .Constants import COUNT

from .Constants import RETAIN
from .Constants import DISCARD

from .Structures import Rank
from .Structures import Individual
from .Structures import ArchiveIndividual
from .Structures import DOEState
from .Structures import Axis
from .Structures import Grid
from .Structures import GridPoint
from .Structures import Issue
from .Structures import Issued
from .Structures import MOEAState

from .Sorting import sort_into_archive

from .Sampling import doe_next
from .Sampling import evolve
from .Sampling import NearExhaustionWarning
from .Sampling import TotalExhaustionError

def create_moea_state(problem, **kwargs):
    """
    problem (Problem): definition of problem structure.

    keywords:
        ranks (int): number of ranks to allocate in the archive
                     (default 100)
        ranksize (int): number of individuals in a rank
                     (default 10,000)
        float_values (RETAIN or DISCARD): what to do with decision
                     variable values.  If RETAIN is selected,
                     the decision variable values will be stored
                     along with the grid points of every individual.
                     If there is a large number of decision variables,
                     this policy results in greatly increased storage
                     requirements, and may require a reduction in
                     ranks or ranksize for the archive to fit in
                     memory.  RETAIN may be a desirable behavior if
                     you are doing local optimization or providing
                     individuals that have been evaluated on a
                     different grid. 
                     If DISCARD is selected, the decision variable
                     values for each individual will not be stored
                     explicitly, and will be regenerated from the
                     individual's grid point.  As long as the
                     individuals provided to the algorithm were
                     evaluated at grid points in decision space, this
                     option is lossless and saves a lot of space.
                     DISCARD is the default for this option.
        random (callable): a real-number generating function,
                     returning a number on the interval [0,1).
                     If none is provided, we fall back on
                     Python's random.random.
        randint (callable): an integer generating function
                     taking two arguments, a lower bound and
                     an (inclusive) upper bound, and returning
                     a number within those bounds.  E.g.
                     calling intrand(0, 1) should return 0 or
                     1.  We expect this to be a random number
                     generator and the algorithm may not
                     converge if it is not.  If not provided,
                     we fall back on Python's random.randint.

    This function creates MOEA state, including
    pre-allocation of a large archive for individuals.

    If the individuals are very large, it may make sense
    to reduce ranks or ranksize to avoid an unnecessary
    allocation.  This entails a tradeoff: fewer ranks save
    memory but risk forgetting that a badly dominated
    point in decision space has already been sampled.
    Smaller ranksize can save a great deal of memory
    if selected appropriately, at the risk of degrading
    algorithmic performance when ranks overflow.
    """
    float_values = kwargs.get("float_values", DISCARD)
    ranks = kwargs.get('ranks', 100)
    ranksize = kwargs.get('ranksize', 10000)
    _random = kwargs.get('random', random)
    _randint = kwargs.get('randint', randint)
    grid = _create_grid(problem.decisions)
    archive = [_empty_rank(problem, float_values, ranksize)
               for _ in range(ranks)]
    rank_A = _empty_rank(problem, float_values, ranksize)
    rank_B = _empty_rank(problem, float_values, ranksize)
    issued = Issued(
        [Issue(grid.GridPoint(*(-1 for _ in problem.decisions)), False)
         for _ in range(ranksize)],
        0,
        set())
    # This is a placeholder.  We call doe() below to
    # initialize the doe state.
    doestate = DOEState(RANDOM, COUNT, 0, 0)

    state = MOEAState(
        problem,
        float_values,
        grid,
        archive,
        set(), # archive_set for Python acceleration
        rank_A,
        rank_B,
        issued,
        _random,
        _randint,
        doestate
    )
    state = doe(state)
    return state

def doe(state, **kwargs):
    """
    Return an MOEAState such that the next generated
    samples will begin to fill out a design of experiments
    on the decision space, rather than doing evolution
    on the archived individuals.  If this function is not
    called, we will do COUNT 100 by default.

    The DOE proceeds as:
        corners         (2 ^ ndv samples)
        center point    (1 sample)
        OFAT            (2 * ndv samples)
        random uniform  (unlimited samples)

    If a different DOE procedure works better for your problem,
    you may substitute one of your choosing by running the
    DOE separately and supplying evaluated individuals to
    the algorithm using return_evaluated_individual before
    beginning the evolution run.

    keywords:
        terminate (CENTERPOINT, OFAT, CORNERS, COUNT):
            indicates stage after which to switch from
            DOE to evolution. Defaults to COUNT 100, which
            means 100 random samples will be generated
            before evolution starts.  If you have a lot
            of decision variables and choose CORNERS,
            you may never finish doing your DOE, but if
            you have a small number of decision variables
            it may be worth while.
            If you specify COUNT, then the number of DOE samples
            is determined by the "count" keyword argument.
        count (int): Number of DOE samples to perform, if COUNT
            is specified as a DOE termination condition.
            Default is ndv.
        stage (CORNERS, CENTERPOINT, OFAT, RANDOM): at which
            stage to start sampling.
    """
    terminate = kwargs.get("terminate", COUNT)
    stage = kwargs.get("stage", RANDOM)
    if terminate == COUNT:
        default_count = len(state.problem.decisions)
        count = kwargs.get("count", default_count)
    else:
        count = 0
    old_doestate = state.doestate
    new_doestate = old_doestate._replace(
        terminate=terminate,
        stage=stage,
        remaining=count)
    new_state = state._replace(doestate=new_doestate)
    return new_state

def return_evaluated_individual(state, individual):
    """
    Return an MOEAState that accounts for the provided
    Individual.
    """
    # produce an ArchiveIndividual from the Individual
    if state.float_values == RETAIN:
        decisions = individual.decisions
    else:
        decisions = tuple()
    grid_point = decisions_to_grid_point(state.grid, individual.decisions)
    archive_set = state.archive_set
    archive_set.add(grid_point)
    state = state._replace(archive_set=archive_set)
    issues = state.issued.issues
    for ii in range(len(issues)):
        issue = issues[ii]
        if issue.outstanding and issue.grid_point == grid_point:
            issues[ii] = issue._replace(outstanding=False)
            issued_set = state.issued.issued_set
            issued_set.remove(grid_point)
            # These _replace calls are not strictly necessary
            # because we're mutating internal structures, but
            # I have a "don't change without calling _replace" rule
            # for myself.
            issued = state.issued._replace(
                issues=issues,
                issued_set=issued_set)
            state = state._replace(issued=issued)
            break

    # ArchiveIndividuals always sort with < and we reverse the transformation
    # when returning Individuals.
    problem = state.problem
    objectives = list()
    for ob, value in zip(problem.objectives, individual.objectives):
        if ob.sense == MINIMIZE:
            objectives.append(value)
        else:
            objectives.append(-value)
    # Ditto constraints
    constraints = list()
    for co, value in zip(problem.constraints, individual.constraints):
        if co.sense == MINIMIZE:
            constraints.append(value)
        else:
            constraints.append(-value)
    archive_individual = ArchiveIndividual(
        True,
        grid_point,
        decisions,
        tuple(objectives),
        tuple(constraints),
        individual.tagalongs)

    # sort the ArchiveIndividual into the archive
    state = sort_into_archive(state, archive_individual)

    # return the state
    return state

def decisions_to_grid_point(grid, decisions):
    """
    grid (Grid): map between the axes and the 
    decisions (tuple of floats): decision variable values

    Returns the GridPoint corresponding to the decisions.
    """
    indices = list()
    for axis, delta, value in zip(grid.axes, grid.deltas, decisions):
        if value <= axis[0]:
            indices.append(0)
        elif value >= axis[-1]:
            indices.append(len(axis) - 1)
        else:
            # return whatever index is closest
            under = int(floor((value - axis[0])/delta))
            under_value = axis[under]
            over_value = axis[under + 1]
            if value - under_value <= over_value - value:
                indices.append(under)
            else:
                indices.append(under + 1)
    return grid.GridPoint(*indices)

def get_iterator(state, rank_number):
    """
    Generator that iterates over the solutions in a rank.
    """
    axes = state.grid.axes
    o_coefficients = list()
    for objective in state.problem.objectives:
        if objective.sense == MAXIMIZE:
            o_coefficients.append(-1)
        else:
            o_coefficients.append(1)
    c_coefficients = list()
    for constraint in state.problem.constraints:
        if constraint.sense == MAXIMIZE:
            c_coefficients.append(-1)
        else:
            c_coefficients.append(1)
    for a_individual in state.archive[rank_number].individuals:
        if a_individual.valid:
            sample = state.grid.Sample(
                *(a[i] for a, i in zip(axes, a_individual.grid_point)))
            objectives = [y * c for y, c
                          in zip(a_individual.objectives, o_coefficients)]
            constraints = [y * c for y, c
                          in zip(a_individual.constraints, c_coefficients)]
            individual = Individual(
                sample,
                objectives,
                constraints,
                a_individual.tagalongs)
            yield individual

def get_sample(state):
    """
    state (MOEAState): current algorithm state

    Returns a new MOEAState and a sample in decision space.
    """
    # Should we do a DOE sample?
    if _should_do_doe(state):
        state, grid_point = doe_next(state)
        # Only DOE samples performed here count against the
        # COUNT termination condition.  (It's perfectly legit
        # for a user to call doe_next to get DOE samples in
        # other circumstances.)
        if state.doestate.terminate == COUNT and state.doestate.remaining > 0:
            state = state._replace(doestate=state.doestate._replace(
                    remaining=state.doestate.remaining - 1))
    else:
        state, grid_point = evolve(state)

    # Add grid_point to issued list
    issues = state.issued.issues
    index = state.issued.index
    issues[index] = Issue(grid_point, True)
    issued_set = state.issued.issued_set
    issued_set.add(grid_point)
    state = state._replace(
        issued=state.issued._replace(
            issues=issues,
            issued_set=issued_set,
            index=(index + 1) % len(issues)
    ))

    grid = state.grid
    sample = grid.Sample(*(a[i] for a, i in zip(grid.axes, grid_point)))
    # Return sample
    return state, sample

def _should_do_doe(state):
    """
    state (MOEAState)
    Returns True if we should do a DOE sample.

    The termination condition is terminate-after-stage,
    so if you specify CENTERPOINT as your termination
    condition, you get one DOE sample.  The way to get
    zero DOE samples is to set the termination condition
    to COUNT and the count to 0.  If the archive is
    completely empty, we're going to issue DOE samples
    anyway, so the user shouldn't be asking for samples
    in that situation if DOE samples are not desired.
    """
    stage = state.doestate.stage
    terminate = state.doestate.terminate
    if terminate == COUNT:
        return state.doestate.remaining > 0
    elif stage == CENTERPOINT:
        return True
    elif stage == OFAT:
        return terminate in (OFAT, CORNERS)
    elif stage == CORNERS:
        return terminate == CORNERS
    elif stage == RANDOM:
        # You're only going to get random points if you're
        # using COUNT for your termination condition.  Or
        # if we've issued a ton of points and gotten none
        # back.
        return False
    # True is safer than False as a fallback, because you
    # can always do DOE but you can't always do selection
    # and variation.  Nevertheless, this return statement
    # should be unreachable.
    return True

def _create_grid(decisions):
    """
    decisions: tuple of Decisions

    Returns a Grid corresponding to the Decisions.
    """
    # Defining a new namedtuple here as a convenience, and
    # mainly so that it's readable in debugging prints.
    _Axes = namedtuple("Axes", (d.name for d in decisions))

    axes = list()
    for decision in decisions:
        decision_range = decision.upper - decision.lower
        number_of_intervals = int(floor(decision_range / decision.delta))
        span = number_of_intervals * decision.delta
        if span + decision.delta - decision_range < 1e-6 * decision.delta:
            # If floor cut us off really close to the upper limit,
            # we need to include it.
            corrected_number_of_intervals = number_of_intervals + 1
            lower = decision.lower
        else:
            # Divide the slop by two and use it as a margin.
            corrected_number_of_intervals = number_of_intervals
            lower = decision.lower + 0.5 * (decision_range - span)
        # accumulate grid values
        values = list()
        value = lower
        for ii in range(corrected_number_of_intervals + 1):
            values.append(value)
            value += decision.delta
        # correct last value to exactly the upper limit because
        # floating point math can accumulate errors
        if values[-1] > decision.upper:
            values[-1] = decision.upper
        axis = Axis(values)
        axes.append(axis)
    _Deltas = namedtuple("Deltas", (d.name for d in decisions))
    grid = Grid(
        _Axes(*axes),
        _Deltas(*(d.delta for d in decisions)),
        namedtuple("GridPoint", (d.name for d in decisions)),
        namedtuple("Sample", (d.name for d in decisions))
    )
    return grid

def _empty_rank(problem, float_values, ranksize):
    # construct bogus individuals to fill the rank
    bogus_grid_point = tuple((999 for _ in problem.decisions))
    if float_values == RETAIN:
        bogus_decisions = tuple((0.0 for _ in problem.decisions))
    else:
        bogus_decisions = tuple()
    bogus_objectives = list()
    # in C99, math.h has macros for infinity and nan
    inf = float("inf")
    ninf = -inf
    # Bogus individuals should never dominate true individuals.
    for objective in problem.objectives:
        if objective.sense == MAXIMIZE:
            bogus_objectives.append(ninf)
        else:
            bogus_objectives.append(inf)
    # Nor should they appear even remotely feasible.
    bogus_constraints = list()
    for constraint in problem.constraints:
        if constraint.sense == MAXIMIZE:
            bogus_constraints.append(ninf)
        else:
            bogus_constraints.append(inf)
    bogus_tagalongs = list((0.0 for _ in problem.tagalongs))
    bogus_archive_individual = ArchiveIndividual(
        False, # and bogus individuals are invalid
        bogus_grid_point,
        bogus_decisions,
        bogus_objectives,
        bogus_constraints,
        bogus_tagalongs)
    return Rank([bogus_archive_individual for _ in range(ranksize)], 0)





"""
Copyright (c) 2018 DecisionVis, LLC. All rights reserved.

Redistribution and use in source and binary forms, with
or without modification, are permitted provided that the
following conditions are met:

1. Redistributions of source code must retain the above
copyright notice, this list of conditions and the following
disclaimer.

2. Redistributions in binary form must reproduce the
above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other
materials provided with the distribution.

3. Neither the name of the copyright holder nor the names
of its contributors may be used to endorse or promote
products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

# Objective / constraint sense
MAXIMIZE = "maximize"
MINIMIZE = "minimize"

# DOE state, termination conditions, and errors
CENTERPOINT = "centerpoint"
OFAT = "OFAT"
CORNERS = "corners"
RANDOM = "random"
EXHAUSTIVE = "exhaustive"   # fell back on grid sweep
EXHAUSTED = "exhausted"     # totally exhausted
COUNT = "count"

# Decision value retention policy
RETAIN = "retain"
DISCARD = "discard"

# Comparison Result
LEFT_DOMINATES = "left dominates"
RIGHT_DOMINATES = "right dominates"
NEITHER_DOMINATES = "neither dominates"



"""
Copyright (c) 2018 DecisionVis, LLC. All rights reserved.

Redistribution and use in source and binary forms, with
or without modification, are permitted provided that the
following conditions are met:

1. Redistributions of source code must retain the above
copyright notice, this list of conditions and the following
disclaimer.

2. Redistributions in binary form must reproduce the
above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other
materials provided with the distribution.

3. Neither the name of the copyright holder nor the names
of its contributors may be used to endorse or promote
products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

from .Constants import MAXIMIZE
from .Constants import MINIMIZE
from .Constants import LEFT_DOMINATES
from .Constants import RIGHT_DOMINATES
from .Constants import NEITHER_DOMINATES

from math import isnan

def sort_into_archive(state, archive_individual):
    archive = state.archive

    rank_A = state.rank_A._replace(occupancy=0)
    rank_B = state.rank_B._replace(occupancy=0)

    # rank A will always be the one we're sorting into
    # the next archive rank, and rank B will be the
    # recipient of displaced individuals.
    rank_A.individuals[0] = archive_individual
    rank_A = rank_A._replace(occupancy=1)

    rank_into = 0
    # loop over archive ranks
    while rank_A.occupancy > 0 and rank_into + 1 < len(archive):
        into = archive[rank_into]
        # print("----")
        # print("before: rank {}".format(rank_into))
        # _print_rank(into)
        # print("before: rank_A")
        # _print_rank(rank_A)
        # print("before: rank_B")
        # _print_rank(rank_B)
        # loop over rank A
        a_remaining = rank_A.occupancy
        for ai, a_ind in enumerate(rank_A.individuals):
            # print("processing {} from rank A: {}".format(ai, a_ind))
            if a_remaining <= 0:
                break
            if not a_ind.valid:
                continue
            a_remaining -= 1
            # loop over rank being sorted into
            i_remaining = into.occupancy
            for ii, i_ind in enumerate(into.individuals):
                if i_remaining <= 0:
                    break
                if not i_ind.valid:
                    continue
                i_remaining -= 1
                dominance = _compare(a_ind, i_ind)
                # invalidate the dominated individual 
                # insert it into rank_B
                # adjust rank occupancy
                if dominance == LEFT_DOMINATES:
                    rank_B, into = move_individual(
                        rank_B, rank_B.occupancy, into, ii)
                elif dominance == RIGHT_DOMINATES:
                    rank_B, rank_A = move_individual(
                        rank_B, rank_B.occupancy, rank_A, ai)
                    # break if rank A individual was dominated and go
                    # to next rank A individual
                    break
        # print("after comparisons: rank {}".format(rank_into))
        # _print_rank(into)
        # print("after comparisons: rank_A")
        # _print_rank(rank_A)
        # print("after comparisons: rank_B")
        # _print_rank(rank_B)
        # insert valid individuals from rank A in "into"
        into, rank_A = fill_rank_from_rank(into, rank_A)

        # print("after fill: rank {}".format(rank_into))
        # _print_rank(into)
        # print("after fill: rank_A")
        # _print_rank(rank_A)
        # print("after fill: rank_B")
        # _print_rank(rank_B)

        # insert overflow in rank B
        rank_B, rank_A = fill_rank_from_rank(rank_B, rank_A)

        # print("after overflow: rank {}".format(rank_into))
        # _print_rank(into)
        # print("after overflow: rank_A")
        # _print_rank(rank_A)
        # print("after overflow: rank_B")
        # _print_rank(rank_B)

        # swap rank A and rank B
        rank_B, rank_A = rank_A, rank_B
        # update the archive
        archive[rank_into] = into

        # print("after: rank {}".format(rank_into))
        # _print_rank(into)
        # print("after: rank_A")
        # _print_rank(rank_A)
        # print("after: rank_B")
        # _print_rank(rank_B)

        rank_into += 1

    # insert all remaining overflow in the last rank
    last_rank = archive[-1]
    last_rank, rank_A = fill_rank_from_rank(last_rank, rank_A)
    archive[-1] = last_rank

    # if there's anything left in rank A, discard the grid points
    # from the archive set
    archive_set = state.archive_set
    for arch_ind in rank_A.individuals:
        if arch_ind.valid:
            archive_set.difference_update((arch_ind.grid_point,))

    state = state._replace(
        rank_A=rank_A,
        rank_B=rank_B,
        archive=archive,
        archive_set=archive_set)

    return state

def _print_rank(rank):
    print("occupancy {}".format(rank.occupancy))
    print("valid {}".format(sum((1 for i in rank.individuals if i.valid))))
    ii = 0
    valid = 0
    while valid < rank.occupancy and ii < len(rank.individuals):
        print(rank.individuals[ii])
        if rank.individuals[ii].valid:
            valid += 1
        ii += 1

def fill_rank_from_rank(destination, source):
    di = 0
    for si, s_ind in enumerate(source.individuals):
        if source.occupancy <= 0:
            break
        if not s_ind.valid:
            continue
        if destination.occupancy >= len(destination.individuals):
            break
        while di < len(destination.individuals):
            if destination.individuals[di].valid:
                di += 1
            else:
                destination, source = move_individual(
                    destination, di, source, si)
                break
        if di >= len(destination.individuals):
            break
    return destination, source

def move_individual(destination_rank, destination_index, source_rank, source_index):
    """
    Move an archive individual from one rank to another rank.
    This function does not check bounds or make any other safety guarantees.
    This function assumes that the individual is valid and adjusts
    occupancy accordingly.

    destination_rank (Rank)
    destination_index (int): index into destination_rank
    source_rank (Rank)
    source_index (int): index into source_rank

    Returns an updated destination rank and source rank.
    """
    individual = source_rank.individuals[source_index]
    destination_rank.individuals[destination_index] = individual
    destination_rank = destination_rank._replace(
        occupancy=destination_rank.occupancy + 1)
    source_rank.individuals[source_index] = individual._replace(
        valid=False)
    source_rank = source_rank._replace(
        occupancy=source_rank.occupancy - 1)
    return destination_rank, source_rank

def _compare(left, right):
    """
    left (Individual)
    right (Individual)
    """
    dleft = True
    dright = True
    # compare 
    for zl, zr in zip(left.constraints, right.constraints):
        if zl <= 0 and zr <= 0:
            # For constraints, we are indifferent to values
            # less than zero.
            continue
        if isnan(zl):
            dleft = False
        if isnan(zr):
            dright = False
        if dleft and zl > zr:
            dleft = False
        elif dright and zr > zl:
            dright = False
        if not (dleft or dright):
            # Early return!! No point in continuing because both
            # individuals violate constraints and neither one
            # dominates the other.
            return NEITHER_DOMINATES
    if dleft and not dright:
        return LEFT_DOMINATES
    if dright and not dleft:
        return RIGHT_DOMINATES

    # Both individuals are feasible, so compare objectives.
    for yl, yr in zip(left.objectives, right.objectives):
        if isnan(yl):
            dleft = False
        if isnan(yr):
            dright = False
        # minimize everything
        if yl > yr:
            dleft = False
        elif yr > yl:
            dright = False
        if not (dleft or dright):
            # early return!!
            return NEITHER_DOMINATES
    if dleft and not dright:
        return LEFT_DOMINATES
    elif dright and not dleft:
        return RIGHT_DOMINATES
    else:
        # if we've oversampled and the individuals are
        # nondominated, arbitrarily say that the
        # "left" individual dominates
        if left.grid_point == right.grid_point:
            return LEFT_DOMINATES
        else:
            return NEITHER_DOMINATES




"""
Copyright (c) 2018 DecisionVis, LLC. All rights reserved.

Redistribution and use in source and binary forms, with
or without modification, are permitted provided that the
following conditions are met:

1. Redistributions of source code must retain the above
copyright notice, this list of conditions and the following
disclaimer.

2. Redistributions in binary form must reproduce the
above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other
materials provided with the distribution.

3. Neither the name of the copyright holder nor the names
of its contributors may be used to endorse or promote
products derived from this software without specific prior
written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

from collections import namedtuple

# Decision: name(string) lower(float) upper(float) delta(float < upper - lower)
Decision = namedtuple("Decision", ("name", "lower", "upper", "delta"))

# Objective: name(string) sense(MINIMIZE or MAXIMIZE)
Objective = namedtuple("Objective", ("name", "sense"))

# Constraint: name(string) sense(MINIMIZE or MAXIMIZE)
# Constraints are special because they are always relative to 0.
# A MINIMIZE constraint is met if <= 0
# A MAXIMIZE constraint is met if >= 0
Constraint = namedtuple("Constraint", ("name", "sense"))

# Tagalong: allow the user to attach other data to an individual
Tagalong = namedtuple("Tagalong", ("name",))

# Problem: everything the library needs to know about the
# optimization problem
Problem = namedtuple("Problem", (
    "decisions",    # tuple of decisions
    "objectives",   # tuple of objectives
    "constraints",  # tuple of constraints
    "tagalongs",    # tuple of tagalongs
))

# ArchiveIndividual: A member of the archive, for internal use.
# It retains the grid position of each individual.  Optionally
# it retains the actual decision variable values as well, but
# the decisions tuple may be zero-length.
# Despite having the same field names as the Problem,
# the values in each of the tuples are just numbers.
ArchiveIndividual = namedtuple("ArchiveIndividual", (
    "valid",        # bool: whether the individual is valid
    "grid_point",   # tuple of indices
    "decisions",    # tuple of floats
    "objectives",   # tuple of floats
    "constraints",  # tuple of floats
    "tagalongs",    # tuple of floats
))

# Individual: An individual from the user's point of view.
# Its decisions are real-valued.
Individual = namedtuple("Individual", (
    "decisions",    # tuple of floats
    "objectives",   # tuple of floats
    "constraints",  # tuple of floats
    "tagalongs",    # tuple of floats
))

# Rank: a set of individuals, none of which dominates any other
# The Individual type itself will be defined at run-time, when
# the Problem is provided to the library.
Rank = namedtuple("Rank", (
    "individuals",  # list of individuals
    "occupancy"     # number of valid individuals present
))

# DOE state: state of the ongoing DOE
DOEState = namedtuple("DOEState", (
    "stage",        # CENTERPOINT, OFAT, CORNERS, RANDOM, EXHAUSTIVE, EXHAUSTED
    "terminate",    # CENTERPOINT, OFAT, CORNERS, COUNT
    "counter",      # int to keep track of where you are in the stage
    "remaining",    # int to keep track of the remaining COUNT
))

# Axis: an axis of the sampling grid is a tuple of decision
# values for a decision
Axis = tuple
# GridPoint: a tuple of indices into the sampling grid
GridPoint = tuple
# Grid: the definition of the grid
Grid = namedtuple("Grid", (
    "axes",         # a tuple of Axes
    "deltas",       # a tuple of floats (to speed conversions)
    "GridPoint",    # a namedtuple type to use for the grid points
    "Sample",       # a namedtuple type to use for samples in decision space
))

# Issue: a GridPoint, and an outstanding flag indicating whether that
# GridPoint is still outstanding.
Issue = namedtuple("Issue", (
    "grid_point",   # a GridPoint
    "outstanding",  # whether the grid point represents an outstanding sample
))

# Issued: rolling record of issued samples
# The issued_set member is a cheat -- it breaks my
# restriction on dynamic data structures.  This is ok
# because we can do without it in the C version -- it's
# just there to accelerate scans that are slow in the
# first place because we're using Python.
Issued = namedtuple("Issued", (
    "issues",       # a list of Issue
    "index",        # where we should write the next sample
    "issued_set",   # set of outstanding grid points (Python acceleration)
))

# Algorithm state at some point in time.
# The archive_set member is a cheat in the same way as
# the issued_set member of Issued is a cheat.  It lets
# us take advantage of Python's fast hash-table lookup
# instead of doing expensive (especially in Python)
# table scans.
MOEAState = namedtuple("MOEAState", (
    "problem",             # a Problem
    "float_values",        # RETAIN or DISCARD floating point decision values
    "grid",                # a Grid
    "archive",             # a list of Ranks
    "archive_set",         # a set of returned grid points for Python acceleration
    "rank_A",              # an extra Rank, needed for sorting
    "rank_B",              # an extra Rank, needed for sorting
    "issued",              # an Issued structure
    "random",              # real-valued [0,1) RNG
    "randint",             # integer-valued [a, b] RNG
    "doestate",            # a DOEState
))



"""
setup.py: Loosely cribbed from the example in pypa

https://github.com/pypa/sampleproject/blob/master/setup.py
"""
from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.rst"), encoding='utf-8') as fp:
    long_description = fp.read()

setup(
    name="deltamoea",
    version='1.5',
    description='Multi-Objective Grid Search Algorithm',
    long_description=long_description,
    url='https://github.com/DecisionVis/deltamoea',
    author="DecisionVis LLC",
    author_email="team@decisionvis.com",
    license="BSD",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7',
    ],
    keywords=[
        "δMOEA", "MOEA", "grid search", "optimization", "multi-objective"],
    packages=['deltamoea'],
)

